"""parser.py

Copyright 2019 Matt Trzcinski

This file is part of siglent_parser.

siglent_parser is free software: you can redistribute it and/or
modify- it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

siglent_parser is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with siglent_parser.  If not, see
<https://www.gnu.org/licenses/>.

"""

import numpy as np
import warnings


class DataType:
    analog = 'analog'
    digital = 'digital'
    analog_digital = 'analog_digital'

# Table 2 T/div Table converted into seconds
TABLE2 =    {
    0  :    1e-009,
    1  :    2e-009,
    2  :    5e-009,
    3  :    1e-008,
    4  :    2e-008,
    5  :    5e-008,
    6  :    1e-007,
    7  :    2e-007,
    8  :    5e-007,
    9  :  0.000001,
    10 :  0.000002,
    11 :  0.000005,
    12 :  0.00001,
    13 :  0.00002,
    14 :  0.00005,
    15 :  0.0001,
    16 :  0.0002,
    17 :  0.0005,
    18 :  0.001,
    19 :  0.002,
    20 :  0.005,
    21 :  0.01,
    22 :  0.02,
    23 :  0.05,
    24 :  0.1,
    25 :  0.2,
    26 :  0.5,
    27 :         1,
    28 :         2,
    29 :         5,
    30 :        10,
    31 :        20,
    32 :        50, }

# Table 3 V/div Table converted to V
TABLE3 = {
    0 : 0.001,
    1 : 0.002,
    2 : 0.005,
    3 :  0.01,
    4 :  0.02,
    5 :  0.05,
    6 :   0.1,
    7 :   0.2,
    8 :   0.5,
    9 :     1,
    10 :    2,
    11 :    5,
    12 :   10,}


class SiglentParser():
    """Siglent SDS1000X/2000X Parser."""

    def __init__(self):

        super().__init__()

        self.file = None
        self.raw_bytes = None

        # https://www.siglentamerica.com/operating-tip/extract-data-binary-file-sds2000x-sds1000x/
        self.wave_length       = None  # 0x00-0x03   Reserved
        self.mso_wave_length   = None  # 0x04-0x07   digital channels wave length
        self.mso_ch_open_num   = None  # 0x10-0x13   32-bit signed integer, little endian. Total enabled digital channels
        self.mso_ch_open_state = []    # 0x14-0x23   8-bit unsigned integer. on/off status of digital channel 0-15 (D0-15), 1 – ON, 0 – OFF
        self.ch1_volt_div_val  = None  # 0xbc-0xbf   32-bit float point, little endian. V/div value of CH1, in units of mV. Such as 2.48 mV/div.
        self.ch2_volt_div_val  = None  # 0xc0-0xc3   32-bit float point, little endian. V/div value of CH2, in units of mV. Such as 2.48 mV/div.
        self.ch3_volt_div_val  = None  # 0xc4-0xc7   32-bit float point, little endian. V/div value of CH3, in units of mV. Such as 2.48 mV/div.
        self.ch4_volt_div_val  = None  # 0xc8-0xcb   32-bit float point, little endian. V/div value of CH4, in units of mV. Such as 2.48 mV/div.
        self.ch1_vert_offset   = None  # 0xdc-0xdf   32-bit signed integer, little endian. Offset value of CH1, with the unit of pixel. Refer to “Calculate the Vertical Offset” to get the actual offset voltage
        self.ch2_vert_offset   = None  # 0xe0-0xe3   Offset value of CH2
        self.ch3_vert_offset   = None  # 0xe4-0xe7   Offset value of CH3
        self.ch4_vert_offset   = None  # 0xe8-0xeb   Offset value of CH4
        self.ch1_volt_div      = None  # 0xf0-0xf3   32-bit signed integer, little endian. The coarse V/div index of CH1. Refer to Table 3 for the details.
        self.ch2_volt_div      = None  # 0xf4-0xf7   The coarse V/div index of CH2.
        self.ch3_volt_div      = None  # 0xf8-0xfb   The coarse V/div index of CH3.
        self.ch4_volt_div      = None  # 0xfc-0xff   The coarse V/div index of CH4.
        self.ch1_on            = None  # 0x100-0x103 32-bit signed integer, little endian. on/off status of CH1, 1 – ON, 0 – OFF
        self.ch2_on            = None  # 0x104-0x107 on/off status of CH2
        self.ch3_on            = None  # 0x108-0x10b on/off status of CH3
        self.ch4_on            = None  # 0x10c-0x10f on/off status of CH4
        self.time_div          = None  # 0x248-0x24b 32-bit signed integer, little endian. T/div index. Refer to Table 2 for the details
        self.time_delay        = None  # 0x250-0x253 32-bit signed integer, little endian. Time delay (Trigger delay) value, in units of pixel. Refer to "Calculate the Time Delay" to get the actual time delay
        self.data              = None  # 0x1470-end  Data. 8-bit unsigned integer for analog data, 1-bit binary integer for digital data, Analog data first, and then digital data. Only data of the enabled channel(s) are stored to the file.

        self.sample_rate = None
        self.time        = None

        self.analog  = None
        self.digital = None

    def read_file(self, file):
        """Read in the raw binary data.

        Args:
            file (str/Path): Raw binary data output from Siglent unit

        Returns:
            bytes: Python bytes object of all the binary data

        """
        self.file = file

        with open(file, "rb") as f:
            self.raw_bytes = f.read()

        return self.raw_bytes

    def parse_file(self, byts):
        """Parse binary data.

        Args:
            byts (bytes): Binary data

        Returns:
            Nothing

        """

        ##### Get header chunks #####
        # self.ba_wave_length       = byts[:0x03 + 1]  # For digital data

        self.ba_mso_wave_length   = byts[0x04:0x07 + 1]
        self.ba_mso_ch_open_num   = byts[0x10:0x13 + 1]
        self.ba_mso_ch_open_state = byts[0x14:0x23 + 1]

        self.ba_ch1_volt_div_val  = byts[0xbc:0xBF + 1]
        self.ba_ch2_volt_div_val  = byts[0xc0:0xC3 + 1]
        self.ba_ch3_volt_div_val  = byts[0xc4:0xC7 + 1]
        self.ba_ch4_volt_div_val  = byts[0xc8:0xCB + 1]

        self.ba_ch1_vert_offset   = byts[0xdc:0xDF + 1]
        self.ba_ch2_vert_offset   = byts[0xe0:0xE3 + 1]
        self.ba_ch3_vert_offset   = byts[0xe4:0xE7 + 1]
        self.ba_ch4_vert_offset   = byts[0xe8:0xEB + 1]

        self.ba_ch1_volt_div      = byts[0xf0:0xF3 + 1]
        self.ba_ch2_volt_div      = byts[0xf4:0xF7 + 1]
        self.ba_ch3_volt_div      = byts[0xf8:0xFB + 1]
        self.ba_ch4_volt_div      = byts[0xfc:0x0FF + 1]

        self.ba_ch1_on            = byts[0x100:0x103 + 1]
        self.ba_ch2_on            = byts[0x104:0x107 + 1]
        self.ba_ch3_on            = byts[0x108:0x10B + 1]
        self.ba_ch4_on            = byts[0x10c:0x10F + 1]

        self.ba_time_div          = byts[0x248:0x24B + 1]
        self.ba_time_delay        = byts[0x250:0x253 + 1]

        ##### Convert to decimal #####

        # Numpy used in order to test for numeric type used (signed,
        # unsigned, etc.)
        # self.wave_length       = np.frombuffer(self.ba_wave_length, dtype='<i4')[0]

        self.mso_wave_length   = np.frombuffer(self.ba_mso_wave_length, dtype='<i4')[0]
        self.mso_ch_open_num   = np.frombuffer(self.ba_mso_ch_open_num, dtype='<i4')[0]
        self.mso_ch_open_state = [np.uint8(b) for b in self.ba_mso_ch_open_state]

        self.ch1_volt_div_val  = np.frombuffer(self.ba_ch1_volt_div_val, dtype='<f4')[0]
        self.ch2_volt_div_val  = np.frombuffer(self.ba_ch2_volt_div_val, dtype='<f4')[0]
        self.ch3_volt_div_val  = np.frombuffer(self.ba_ch3_volt_div_val, dtype='<f4')[0]
        self.ch4_volt_div_val  = np.frombuffer(self.ba_ch4_volt_div_val, dtype='<f4')[0]

        self.ch1_vert_offset   = np.frombuffer(self.ba_ch1_vert_offset, dtype='<i4')[0]
        self.ch2_vert_offset   = np.frombuffer(self.ba_ch2_vert_offset, dtype='<i4')[0]
        self.ch3_vert_offset   = np.frombuffer(self.ba_ch3_vert_offset, dtype='<i4')[0]
        self.ch4_vert_offset   = np.frombuffer(self.ba_ch4_vert_offset, dtype='<i4')[0]

        self.ch1_volt_div      = np.frombuffer(self.ba_ch1_volt_div, dtype='<i4')[0]
        self.ch2_volt_div      = np.frombuffer(self.ba_ch2_volt_div, dtype='<i4')[0]
        self.ch3_volt_div      = np.frombuffer(self.ba_ch3_volt_div, dtype='<i4')[0]
        self.ch4_volt_div      = np.frombuffer(self.ba_ch4_volt_div, dtype='<i4')[0]

        self.ch1_on            = np.frombuffer(self.ba_ch1_on, dtype='<i4')[0]
        self.ch2_on            = np.frombuffer(self.ba_ch2_on, dtype='<i4')[0]
        self.ch3_on            = np.frombuffer(self.ba_ch3_on, dtype='<i4')[0]
        self.ch4_on            = np.frombuffer(self.ba_ch4_on, dtype='<i4')[0]

        self.time_div          = np.frombuffer(self.ba_time_div, dtype='<i4')[0]
        self.time_delay        = np.frombuffer(self.ba_time_delay, dtype='<i4')[0]

        self.data = np.frombuffer(byts[0x1470:], dtype='<u1')

        # The calculations section of the specification is not
        # particularly clear (or consistent).  These variables were
        # obtained through trial and error, comparing against the
        # "Para Save" CSV output.

        # Consider 'div' to be a divisor and 'div_val' to be a
        # dividend; perform a scaling on the divisor
        self.real_ch1_volt_div_val = self.ch1_volt_div_val / (TABLE3[self.ch1_volt_div] * 1e6)
        self.real_ch1_vert_offset  = self.calc_vertical_offset(self.ch1_vert_offset,
                                                               self.real_ch1_volt_div_val)

        self.real_ch2_volt_div_val = self.ch2_volt_div_val / (TABLE3[self.ch2_volt_div] * 1e6)
        self.real_ch2_vert_offset  = self.calc_vertical_offset(self.ch2_vert_offset,
                                                               self.real_ch2_volt_div_val)

        self.real_ch3_volt_div_val = self.ch3_volt_div_val / (TABLE3[self.ch3_volt_div] * 1e6)
        self.real_ch3_vert_offset  = self.calc_vertical_offset(self.ch3_vert_offset,
                                                               self.real_ch3_volt_div_val)

        self.real_ch4_volt_div_val = self.ch4_volt_div_val / (TABLE3[self.ch4_volt_div] * 1e6)
        self.real_ch4_vert_offset  = self.calc_vertical_offset(self.ch4_vert_offset,
                                                               self.real_ch4_volt_div_val)

        self.data_type = self.get_data_type()

        try:
            if self.data_type == DataType.analog:

                # According to "Calculate the Sample Rate",
                # wave_length is either the length of each frame or
                # the length of the data section.  No type given in
                # specification; set to '<i4' to match other 32-bit
                # chunks.
                self.wave_length = np.int32(len(self.data))

                self.sample_rate = self.wave_length / (14 * TABLE2[self.time_div])

                self.analog = self.calc_data_to_volt(self.data,
                                                     self.real_ch1_volt_div_val,
                                                     self.real_ch1_vert_offset)

                self.time = np.arange(len(self.analog)) / self.sample_rate

            else:  # Contains digital data
                warnings.warn("Parsing of digital data not currently implemented.")

        except:
            warnings.warn("Unable to identify data as analog or digital.")
            pass

    def calc_vertical_offset(self, ch_vert_offset, ch_volt_div_val, pixels_per_div=50):
        """Calculate the vertical offset.

        The comments state, "pixels_per_div = 50 # total display
        pixels in a vertical division, on SDS2000X is 50"

        """
        return (ch_vert_offset - 220) * (ch_volt_div_val / pixels_per_div)

    def calc_data_to_volt(self, data, ch_volt_div_val, ch_vert_offset, code_per_div=25):
        """Convert the data to voltage.

        The comments state, "code_per_div = 50 # total data code in a
        horizontal division, on SDS2000X is 25"

        """
        try:
            data = data.astype(np.float32)
        except:
            pass
        return (data - 128) * ch_volt_div_val / 1000 / code_per_div + ch_vert_offset

    def parse_binary(self, file):
        """Read in binary Siglent data and return well-formatted numpy array.

        Args:
            file (str/Path): Path to binary file

        Returns:
            numpy.array: Parsed data

        """

        byts = self.read_file(file)
        self.parse_file(byts)

        my_descr       = [('time', '<f8'), ('volts', '<f4')]
        my_df          = np.empty(shape=len(self.data), dtype=my_descr)

        my_df['time']  = self.time
        my_df['volts'] = self.analog

        return my_df

    def get_data_type(self):
        """Return whether the data is analog-only, digital-only, or
        analog-digital.

        Returns:
            DataType (str): String describing data type

        """
        if self.mso_ch_open_num == 0:
            return DataType.analog
        elif self.wave_length == self.mso_wave_length:
            return DataType.analog_digital
        else:
            return DataType.digital
