"""test_siglent_parser.py

Run from top project directory: ./siglent_parser

Best used with PyTest, although PyTtest is not required to run tests.
To run a test without PyTest, call the tests you want to run main().

Run all tests with PyTest:

    python -m pytest test_siglent_parser.py

Run a specific test with PyTest:

    python -m pytest test_siglent_parser.py::test_class::test_method

You may see warnings that look like:

  c:/projects/siglent_parser/siglent_parser/siglent_parser.py:77: FutureWarning:
    elementwise comparison failed; returning scalar instead, but in
    the future will perform elementwise comparison

Ignore these.  They exist because Numpy and Python are in a standoff
about how to handle comparisons between numpy numeric types and Python
native types.  See https://stackoverflow.com/a/46721064.

Copyright 2019 Matt Trzcinski

This file is part of siglent_parser.

siglent_parser is free software: you can redistribute it and/or
modify- it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

siglent_parser is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with siglent_parser.  If not, see
<https://www.gnu.org/licenses/>.

"""

import math
import numpy as np
from pathlib import Path
from unittest import mock
from siglent_parser import SiglentParser
from siglent_parser.parser import TABLE3
from siglent_parser.parser import TABLE2


class TestConstants():

    ##################################################
    # pathing
    ##################################################
    project_dir = Path.cwd()

    ##################################################
    # float tolerance
    ##################################################

    # Tolerance for floating point comparison
    tolerance = 1e-10

    ##################################################
    # mock data
    ##################################################

    # Repeated sequency of 0x0 to 0xff 300 bytes long; data starts at
    # 0x1470 which is index 5232
    mock_data = bytearray([i % 255 for i in range(5532)])

    # Trick the mock data into being DataType.analog
    mock_data[0x10:0x14] = b'\x00\x00\x00\x00'
    mock_data = bytes(mock_data)

    # The mock data values are not realistic and thus not found in the
    # TABLE2/3 lookup, causing an error in variable calculations.
    TABLE2[1296845642] = 11  # self.time_div
    TABLE3[-202182160] = 23  # self.ch1_volt_div
    TABLE3[-134810124] = 58  # self.ch2_volt_div
    TABLE3[-67438088]  = 13  # self.ch3_volt_div
    TABLE3[16711164]   = 21  # self.ch4_volt_div

    ##################################################
    # source files
    ##################################################

    # From CSV with "Para Save" option ON
    # |------------------+---------------------|
    # | Record Length    | 140000              |
    # | Sample Interval  | CH1:0.0000001000000 |
    # | Vertical Units   | CH1:V               |
    # | Vertical Scale   | CH1:1.00            |
    # | Vertical Offset  | CH1:-2.86000        |
    # | Horizontal Units | s                   |
    # | Horizontal Scale | 0.0010000001        |
    # | Model Number     | SDS2304X            |
    # | Serial Number    | SDS2XJBD1R2760      |
    # | Software Version | 1.2.2.2 R15         |
    # |------------------+---------------------|
    analog_sample         = project_dir.joinpath(r"tests\resources\analog_sample2\SDS00001.BIN")

    # These files are known only to be digital or analog-digital.  No
    # CSVs with "Para Save" were output during creation.
    digital_sample        = project_dir.joinpath(r"tests\resources\digital_sample1\SDS00002.BIN")
    analog_digital_sample = project_dir.joinpath(r"tests\resources\digital_analog_sample1\SDS00003.BIN")


class TestSiglentParser(TestConstants):

    ##################################################
    # test_read_file
    ##################################################
    def test_read_file_real(self):
        """Test that file is read correctly, using real Siglent data."""

        sp = SiglentParser()
        raw_sample2 = sp.read_file(self.analog_sample)
        assert len(raw_sample2) == 145232

    def test_read_file_mock(self):
        """Test that file is read in correctly, using mock data."""

        sp = SiglentParser()
        with mock.patch('builtins.open', mock.mock_open(read_data=self.mock_data)):
            raw_mock = sp.read_file('mock_data.BIN')

        assert len(raw_mock) == 5532

    ##################################################
    # test_parse_file
    ##################################################
    def test_parse_file_type(self):
        """Test that parse_file reads chunk of correct type and size."""

        sp = SiglentParser()
        sp.parse_file(self.mock_data)

        assert sp.mso_wave_length.dtype == np.dtype('<i4')
        assert sp.mso_ch_open_num.dtype == np.dtype('<i4')

        for mcos in range(15):
            assert sp.mso_ch_open_state[mcos].dtype == np.uint8

        assert sp.ch1_volt_div_val.dtype == np.dtype('<f4')
        assert sp.ch2_volt_div_val.dtype == np.dtype('<f4')
        assert sp.ch3_volt_div_val.dtype == np.dtype('<f4')
        assert sp.ch4_volt_div_val.dtype == np.dtype('<f4')

        assert sp.ch1_vert_offset.dtype == np.dtype('<i4')
        assert sp.ch2_vert_offset.dtype == np.dtype('<i4')
        assert sp.ch3_vert_offset.dtype == np.dtype('<i4')
        assert sp.ch4_vert_offset.dtype == np.dtype('<i4')

        assert sp.ch1_volt_div.dtype == np.dtype('<i4')
        assert sp.ch2_volt_div.dtype == np.dtype('<i4')
        assert sp.ch3_volt_div.dtype == np.dtype('<i4')
        assert sp.ch4_volt_div.dtype == np.dtype('<i4')

        assert sp.ch1_on.dtype == np.dtype('<i4')
        assert sp.ch2_on.dtype == np.dtype('<i4')
        assert sp.ch3_on.dtype == np.dtype('<i4')
        assert sp.ch4_on.dtype == np.dtype('<i4')

        assert sp.time_div.dtype   == np.dtype('<i4')
        assert sp.time_delay.dtype == np.dtype('<i4')

        assert sp.data_type == 'analog'

        assert sp.data.dtype        == np.dtype('<u1')
        assert sp.wave_length.dtype == np.dtype('<i4')
        assert sp.time.dtype        == np.dtype('<f8')  # Division forces float64 type
        assert sp.analog.dtype      == np.dtype('<f4')

    def test_parse_file_value(self):
        """Test that parse_file handles parsing correctly."""

        sp = SiglentParser()
        sp.parse_file(self.mock_data)

        # Assertion values obtained using the appropriate endianness/
        # integer type and one of the following:
        #
        # Int:   int.from_bytes(mock_data[0x10:0x14], 'little', signed=True)
        # Float: struct.unpack('f', mock_data[0xbc:0xc0])
        assert sp.mso_wave_length       == 117835012  # mock_data[0x04:0x08] -> b'\x04\x05\x06\x07'
        assert sp.mso_ch_open_num       == 0          # mock_data[0x10:0x14] -> b'\x00\x00\x00\x00', manually replaced
        assert sp.mso_ch_open_state[0]  == 20         # mock_data[0x14]      -> b'\x14'
        assert sp.mso_ch_open_state[1]  == 21         # mock_data[0x15]      -> b'\x15'
        assert sp.mso_ch_open_state[2]  == 22         # mock_data[0x16]      -> b'\x16'
        assert sp.mso_ch_open_state[3]  == 23         # mock_data[0x17]      -> b'\x17'
        assert sp.mso_ch_open_state[4]  == 24         # mock_data[0x18]      -> b'\x18'
        assert sp.mso_ch_open_state[5]  == 25         # mock_data[0x19]      -> b'\x19'
        assert sp.mso_ch_open_state[6]  == 26         # mock_data[0x20]      -> b'\x20'
        assert sp.mso_ch_open_state[7]  == 27         # mock_data[0x21]      -> b'\x21'
        assert sp.mso_ch_open_state[8]  == 28         # mock_data[0x22]      -> b'\x22'
        assert sp.mso_ch_open_state[9]  == 29         # mock_data[0x23]      -> b'\x23'
        assert sp.mso_ch_open_state[10] == 30         # mock_data[0x24]      -> b'\x24'
        assert sp.mso_ch_open_state[11] == 31         # mock_data[0x25]      -> b'\x25'
        assert sp.mso_ch_open_state[12] == 32         # mock_data[0x26]      -> b'\x26'
        assert sp.mso_ch_open_state[13] == 33         # mock_data[0x27]      -> b'\x27'
        assert sp.mso_ch_open_state[14] == 34         # mock_data[0x28]      -> b'\x28'
        assert sp.mso_ch_open_state[15] == 35         # mock_data[0x29]      -> b'\x29'

        assert math.isclose(sp.ch1_volt_div_val, -1.4901652336 , rel_tol=self.tolerance)  # mock_data[0xbc:0xc0] -> b'\xbc\xbd\xbe\xbf'
        assert math.isclose(sp.ch2_volt_div_val, -389.513671875, rel_tol=self.tolerance)  # mock_data[0xc0:0xc4] -> b'\xc0\xc1\xc2\xc3'
        assert math.isclose(sp.ch3_volt_div_val, -101771.53125 , rel_tol=self.tolerance)  # mock_data[0xc4:0xc8] -> b'\xc4\xc5\xc6\xc7'
        assert math.isclose(sp.ch4_volt_div_val, -26579856.0   , rel_tol=self.tolerance)  # mock_data[0xc8:0xcc] -> b'\xc8\xc9\xca\xcb'

        assert sp.ch1_vert_offset == -539042340  # mock_data[0xdc:0xe0] -> b'\xdc\xdd\xde\xdf'
        assert sp.ch2_vert_offset == -471670304  # mock_data[0xe0:0xe4] -> b'\xe0\xe1\xe2\xe3'
        assert sp.ch3_vert_offset == -404298268  # mock_data[0xe4:0xe8] -> b'\xe4\xe5\xe6\xe7'
        assert sp.ch4_vert_offset == -336926232  # mock_data[0xe8:0xec] -> b'\xe8\xe9\xea\xeb'

        assert sp.ch1_volt_div == -202182160     # mock_data[0xf0:0xf4]  -> b'\xf0\xf1\xf2\xf3'
        assert sp.ch2_volt_div == -134810124     # mock_data[0xf4:0xf8]  -> b'\xf4\xf5\xf6\xf7'
        assert sp.ch3_volt_div == -67438088      # mock_data[0xf8:0xfc]  -> b'\xf8\xf9\xfa\xfb'
        assert sp.ch4_volt_div == 16711164       # mock_data[0xfc:0x100] -> b'\xfc\xfd\xfe\x00'

        assert sp.ch1_on == 67305985             # mock_data[0x100:0x104] -> b'\x01\x02\x03\x04'
        assert sp.ch2_on == 134678021            # mock_data[0x104:0x108] -> b'\x05\x06\x07\x08'
        assert sp.ch3_on == 202050057            # mock_data[0x108:0x10c] -> b'\t\n\x0b\x0c'
        assert sp.ch4_on == 269422093            # mock_data[0x10c:0x110] -> b'\r\x0e\x0f\x10'

        assert sp.time_div == 1296845642         # mock_data[0x248:0x24c] -> b'JKLM'
        assert sp.time_delay == 1431589714       # mock_data[0x250:0x254] -> b'RSTU'

        assert sp.wave_length == 300             # mock_data is 5532, data starts at 0x1470=5232

        assert math.isclose(sp.sample_rate, 1.948051948051948, rel_tol=self.tolerance)  # 300 / (14 * 11)

    ##################################################
    # test_calc_vertical_offset
    ##################################################
    def test_calc_vertical_offset(self):
        """Test that the return value matches the example given in the
        specification.

        """

        sp = SiglentParser()
        pixels_per_div = 50
        ch_vert_offset = 270
        ch_volt_div_val = 50
        assert sp.calc_vertical_offset(ch_vert_offset, ch_volt_div_val, pixels_per_div) == 50

    ##################################################
    # test_calc_data_to_volt
    ##################################################
    def test_calc_data_to_volt(self):
        """Test that the return value matches the example given in the
        specification.

        """

        sp = SiglentParser()
        data = 194
        ch_volt_div_val = 5000
        ch_vert_offset = -7.7
        code_per_div = 25

        rv = sp.calc_data_to_volt(data, ch_volt_div_val, ch_vert_offset, code_per_div)
        assert math.isclose(rv, 5.5, rel_tol=self.tolerance)

    def test_calc_data_to_volt_return_type(self):
        """Test method return type."""

        sp = SiglentParser()
        data = 194
        ch_volt_div_val = 5000
        ch_vert_offset = -7.7
        code_per_div = 25

        rv = sp.calc_data_to_volt(data, ch_volt_div_val, ch_vert_offset, code_per_div)
        assert isinstance(rv, np.float) is True

    ##################################################
    # test_get_data_type
    ##################################################
    def test_get_data_type(self):
        """Test that correct data is identified."""

        sp = SiglentParser()

        # Analog
        sp.mso_ch_open_num = 0
        assert sp.get_data_type() == 'analog'

        # Analog-Digital
        sp.mso_ch_open_num = 16    # is non-zero
        sp.wave_length     = 7000
        sp.mso_wave_length = 7000  # wave lengths same
        assert sp.get_data_type() == 'analog_digital'

        # Digital
        sp.mso_ch_open_num = 16    # is non-zero
        sp.wave_length     = 7000000
        sp.mso_wave_length = 7000  # wave lengths not same
        assert sp.get_data_type() == 'digital'

    ##################################################
    # test_parse_binary
    ##################################################
    def test_parse_binary(self):
        """Test return value of binary parser."""

        sp = SiglentParser()

        with mock.patch('builtins.open', mock.mock_open(read_data=self.mock_data)):
            rv = sp.parse_binary('mock_data.BIN')

        assert isinstance(rv, np.ndarray)
        assert rv.dtype == [('time', '<f8'), ('volts', '<f4')]


def main():
    """Main function used to run tests manually.

    Use PyTest to run tests in bulk.

    """

    # tc = TestClass
    # TestClass.test_thing()
    pass


if __name__ == '__main__':

    import time

    start_time = time.time()

    main()

    print("\nThe chosen tests have all passed.")
    print("--- %s seconds ---" % (time.time() - start_time))
